#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "driver/gpio.h"
#include "esp_mac.h"
#include "esp_wifi.h"
#include "esp_event.h"
#include "esp_log.h"
#include "nvs_flash.h"

#include "lwip/err.h"
#include "lwip/sys.h"

#include "wifi.h"

// event queue for the ISR passing events to the task
static QueueHandle_t gpio_evt_queue = NULL;

#define CFG_WIFI_SSID      "BSS-Test"       // SSID for the softAP / station
#define CFG_WIFI_PASS      "hesloheslo"     // password for the SSID
#define CFG_WIFI_CHANNEL   4                // softAP: start up the softAP at this channel
#define CFG_MAX_STA_CONN   10               // softAP: maximum number of stations connected at a single time

#define GPIO_WIFI_SWITCH_PIN    27          // on-board switch
#define GPIO_LED_PIN            2           // on-board LED

#define ESP_INTR_FLAG_DEFAULT   0           // interrupt flag to be reset

static void IRAM_ATTR gpio_isr_handler(void* arg)
{
    // take the GPIO number and send the result into event queue for the task to handle
    uint32_t gpio_num = (uint32_t) arg;
    xQueueSendFromISR(gpio_evt_queue, &gpio_num, NULL);
}

static void switch_wifi_state(bool ap_mode) {

    // if there is STA connecting routine in progress, cancel it
    if (wifi_current_sta_state == WIFI_STA_STATE_CONNECTING) {
        wifi_current_sta_state = WIFI_STA_STATE_CANCELLING;

        // wait until it's shut down
        while (wifi_current_sta_state != WIFI_STA_STATE_NONE) {
            vTaskDelay(1000 / portTICK_PERIOD_MS);
        }
    }

    // just for us to know, when to de-initialize
    static bool initialized_before = false;

    // de-initialize WiFi if needed
    if (initialized_before) {
        esp_wifi_stop();
        esp_wifi_deinit();
    }

    initialized_before = true;

    // start up in given mode
    if (ap_mode) {
        wifi_init_softap(CFG_WIFI_SSID, CFG_WIFI_PASS, CFG_WIFI_CHANNEL, CFG_MAX_STA_CONN);
    }
    else {
        wifi_init_station(CFG_WIFI_SSID, CFG_WIFI_PASS);
    }
}

static void wifi_task_main() {
    
    bool is_ap = false;

    // at first, intialize as station
    switch_wifi_state(is_ap);

    // we will just wait for incoming GPIO interrupts here - after the GPIO interrupt comes, switch from AP to STA or vice versa
    uint32_t io_num;
    for (;;) {
        if (xQueueReceive(gpio_evt_queue, &io_num, portMAX_DELAY)) {

            int level = gpio_get_level(io_num);
            printf("GPIO[%"PRIu32"] interrupt detected, value=%d\n", io_num, level);
            // positive edge - switch between modes
            if (level == 1) {
                is_ap = !is_ap;
                switch_wifi_state(is_ap);
            }
        }
    }
}

void app_main(void)
{
    // when we use WiFi, we most likely want to initialize non-volatile storage (flash) to
    // store some settings to NVS. For now, let's just prepare it here, so we have it available
    // later when we actually need it (e.g., when putting the ESP to sleep)
    esp_err_t ret = nvs_flash_init();
    if (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND) {
        ESP_ERROR_CHECK(nvs_flash_erase());
        ret = nvs_flash_init();
    }
    ESP_ERROR_CHECK(ret);

    // output only errors
    // we may want to set it to info/verbose if we want to see more
    esp_log_level_set("*", ESP_LOG_ERROR);

    // initialize TCP/IP stack (load to memory)
    esp_netif_init();
    // create default event loop (for networking subsystem)
    esp_event_loop_create_default();

    // hook our custom event handlers
    wifi_init_event_handlers();

    printf("BSS project starting\n");

    // create event queue for the GPIO interrupts
    gpio_evt_queue = xQueueCreate(10, sizeof(uint32_t));

    gpio_config_t io_conf = {};

    // set up input pins
    io_conf.intr_type = GPIO_INTR_POSEDGE;                  // trigger interrupt on positive edge
    io_conf.pin_bit_mask = (1ULL << GPIO_WIFI_SWITCH_PIN);  // bit mask of the pins - use pin 27 (on-board switch)
    io_conf.mode = GPIO_MODE_INPUT;                         // set pin to input mode
    io_conf.pull_up_en = 1;                                 // enable pull-up mode ("default" state is 1)
    gpio_config(&io_conf);

    // set up output pins
    io_conf.intr_type = GPIO_INTR_DISABLE;                  // disable interrupts
    io_conf.pin_bit_mask = (1ULL << GPIO_LED_PIN);          // bit mask of the pins - use pin 2 (on-board LED)
    io_conf.mode = GPIO_MODE_OUTPUT;                        // set pin to output mode
    io_conf.pull_down_en = 0;                               // disable pull-down
    io_conf.pull_up_en = 0;                                 // disable pull-up
    gpio_config(&io_conf);

    // enable interrupts, install ISR handler
    gpio_set_intr_type(GPIO_WIFI_SWITCH_PIN, GPIO_INTR_ANYEDGE);
    gpio_install_isr_service(ESP_INTR_FLAG_DEFAULT);
    gpio_isr_handler_add(GPIO_WIFI_SWITCH_PIN, gpio_isr_handler, (void*)GPIO_WIFI_SWITCH_PIN);

    // create wifi switcher task
    xTaskCreate(wifi_task_main, "wifi_task_main", 4096, NULL, 10, NULL);

    // in this task, just blink some LEDs
    while(1) {
        gpio_set_level(GPIO_LED_PIN, true);
        vTaskDelay(1000 / portTICK_PERIOD_MS);
        gpio_set_level(GPIO_LED_PIN, false);
        vTaskDelay(1000 / portTICK_PERIOD_MS);
    }
}
