#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/event_groups.h"
#include "driver/gpio.h"
#include "esp_mac.h"
#include "esp_wifi.h"
#include "esp_event.h"
#include "esp_log.h"
#include "nvs_flash.h"

#include "wifi.h"

#define WIFI_CONNECTED_BIT BIT0
#define WIFI_FAIL_BIT      BIT1

static EventGroupHandle_t s_wifi_event_group;
static esp_netif_t* s_wifi_netif = NULL;

wifi_sta_state_t wifi_current_sta_state;

static void wifi_event_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data) {
    if (event_id == WIFI_EVENT_AP_STACONNECTED) {
        wifi_event_ap_staconnected_t* event = (wifi_event_ap_staconnected_t*) event_data;
        printf("SoftAP: station "MACSTR" joined, AID=%d\n", MAC2STR(event->mac), event->aid);
    } else if (event_id == WIFI_EVENT_AP_STADISCONNECTED) {
        wifi_event_ap_stadisconnected_t* event = (wifi_event_ap_stadisconnected_t*) event_data;
        printf("SoftAP: station "MACSTR" left, AID=%d\n", MAC2STR(event->mac), event->aid);
    } else if (event_id == WIFI_EVENT_STA_START) {
        esp_wifi_connect();
    } else if (event_id == WIFI_EVENT_STA_DISCONNECTED) {

        wifi_event_sta_disconnected_t* evt = (wifi_event_sta_disconnected_t*)event_data;

        if (evt->reason == WIFI_REASON_ASSOC_LEAVE) {
            printf("Association terminated, you probably terminated the STA mode\n");
        }
        else {
            printf("Connection attempt... disconnect reason: %d\n", evt->reason);
        }

        if (wifi_current_sta_state == WIFI_STA_STATE_CANCELLING) {
            printf("Cancelling connection...\n");
            wifi_current_sta_state = WIFI_STA_STATE_NONE;
            xEventGroupSetBits(s_wifi_event_group, WIFI_FAIL_BIT);
            return;
        }

        esp_wifi_connect();
    }
}

static void ip_event_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data) {
    if (event_id == IP_EVENT_STA_GOT_IP) {
        ip_event_got_ip_t* event = (ip_event_got_ip_t*) event_data;
        printf("Station connected, obtained IP:" IPSTR "\n", IP2STR(&event->ip_info.ip));

        wifi_current_sta_state = WIFI_STA_STATE_CONNECTED;
        xEventGroupSetBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
    }
}

void wifi_init_event_handlers() {
    esp_event_handler_instance_register(WIFI_EVENT, ESP_EVENT_ANY_ID, &wifi_event_handler, NULL, NULL);
    esp_event_handler_instance_register(IP_EVENT, IP_EVENT_STA_GOT_IP, &ip_event_handler, NULL, NULL);

    s_wifi_event_group = xEventGroupCreate();
}

void wifi_init_softap(const char* ssid, const char* password, int channel, int max_connections) {

    printf(">> Starting softAP\n");

    wifi_current_sta_state = WIFI_STA_STATE_NONE;

    if (s_wifi_netif != NULL)
        esp_netif_destroy_default_wifi(s_wifi_netif);

    s_wifi_netif = esp_netif_create_default_wifi_ap();

    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    esp_wifi_init(&cfg);

    wifi_config_t wifi_config;

    const int ssidLen = strlen(ssid);
    const int passwordLen = strlen(password);
    
    strncpy((char*)wifi_config.ap.ssid, ssid, ssidLen > 32 ? 32 : ssidLen);
    wifi_config.ap.ssid_len = ssidLen;
    strncpy((char*)wifi_config.ap.password, password, passwordLen > 64 ? 64 : passwordLen);
    wifi_config.ap.channel = channel;
    wifi_config.ap.max_connection = max_connections;
    wifi_config.ap.authmode = (passwordLen == 0) ? WIFI_AUTH_OPEN : WIFI_AUTH_WPA2_PSK;
    wifi_config.ap.pmf_cfg.required = true;

    esp_wifi_set_mode(WIFI_MODE_AP);
    esp_wifi_set_config(WIFI_IF_AP, &wifi_config);
    esp_wifi_start();

    printf("SoftAP initialized: SSID=%s, password=%s, channel=%d\n", ssid, password, channel);
}

void wifi_init_station(const char* ssid, const char* password) {

    printf(">> Starting station\n");

    wifi_current_sta_state = WIFI_STA_STATE_CONNECTING;

    if (s_wifi_netif != NULL)
        esp_netif_destroy_default_wifi(s_wifi_netif);

    s_wifi_netif = esp_netif_create_default_wifi_sta();

    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    esp_wifi_init(&cfg);

    wifi_config_t wifi_config = {
        .sta = {
            .ssid = "",
            .password = "",
        },
    };

    strcpy((char*)wifi_config.sta.ssid, ssid);
    strcpy((char*)wifi_config.sta.password, password);

    esp_wifi_set_mode(WIFI_MODE_STA);
    esp_wifi_set_config(WIFI_IF_STA, &wifi_config);
    esp_wifi_start();

    printf("Station trying to connect now...\n");
}
