#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include "freertos/FreeRTOS.h"
#include "freertos/semphr.h"
#include "freertos/timers.h"
#include "nvs_flash.h"
#include "esp_random.h"
#include "esp_event.h"
#include "esp_netif.h"
#include "esp_wifi.h"
#include "esp_log.h"
#include "esp_mac.h"
#include "esp_now.h"
#include "esp_crc.h"

#include "espnow_layer.h"
#include "linkproto.h"

#include "esp_log.h"

#define TAG "ESPNOW"
// master key to be user (as primary and local) - must be 16 bytes long
#define MASTER_KEY "key1234567890123"

#define USE_WIFI_CHANNEL 4 // TODO: change it to configurable

const uint8_t Broadcast_L2_Address[ESP_NOW_ETH_ALEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

namespace {
    CESP_NOW* cb_instance = nullptr;

    void espnow_send_cb(const uint8_t *mac_addr, esp_now_send_status_t status) {
        if (cb_instance)
            cb_instance->Send_Callback(mac_addr, status);
    }

    void espnow_recv_cb(const esp_now_recv_info_t *recv_info, const uint8_t *data, int len) {
        if (cb_instance)
            cb_instance->Receive_Callback(recv_info, data, len);
    }
}

CESP_NOW::CESP_NOW() {
    //
}

bool CESP_NOW::Initialize() {
    return Initialize_WiFi() && Initialize_ESP_Now();
}

bool CESP_NOW::Initialize_WiFi() {
    ESP_LOGI(TAG, "Initializing WiFi");

    esp_netif_init();
    esp_event_loop_create_default();
    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    esp_wifi_init(&cfg);
    esp_wifi_set_storage(WIFI_STORAGE_RAM);
    esp_wifi_set_mode(WIFI_MODE_AP);
    esp_wifi_start();
    esp_wifi_set_channel(USE_WIFI_CHANNEL, WIFI_SECOND_CHAN_NONE);

    // for long-range communication, we may use something like this:
    //esp_wifi_set_protocol(ESPNOW_WIFI_IF, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N | WIFI_PROTOCOL_LR);

    return true;
}

bool CESP_NOW::Initialize_ESP_Now() {
    ESP_LOGI(TAG, "Initializing ESP-NOW");

    esp_now_init();

    cb_instance = this;
    esp_now_register_send_cb(&espnow_send_cb);
    esp_now_register_recv_cb(&espnow_recv_cb);

    // to save power, we may want to set the wake window and wake interval (basically to implement asynchronous TDMA)
    //esp_now_set_wake_window(CONFIG_ESPNOW_WAKE_WINDOW);
    //esp_wifi_connectionless_module_set_wake_interval(CONFIG_ESPNOW_WAKE_INTERVAL);

    // set primary master key
    esp_now_set_pmk((uint8_t*)MASTER_KEY);

    esp_now_peer_info_t* peer = (esp_now_peer_info_t*)malloc(sizeof(esp_now_peer_info_t));
    if (peer == NULL) {
        esp_now_deinit();
        return false;
    }

    Add_Peer(Broadcast_L2_Address);

    return true;
}

bool CESP_NOW::Add_Peer(const uint8_t* dst_mac) {

    esp_now_peer_info_t* peer = (esp_now_peer_info_t*)malloc(sizeof(esp_now_peer_info_t));
    if (peer == NULL) {
        esp_now_deinit();
        return false;
    }

    memset(peer, 0, sizeof(esp_now_peer_info_t));
    peer->channel = USE_WIFI_CHANNEL;
    peer->ifidx = (wifi_interface_t)ESP_IF_WIFI_AP;
    peer->encrypt = false;
    memcpy(peer->peer_addr, dst_mac, ESP_NOW_ETH_ALEN);
    esp_now_add_peer(peer);
    free(peer);

    return true;
}

bool CESP_NOW::Remove_Peer(const uint8_t* dst_mac) {
    esp_now_del_peer(dst_mac);

    return true;
}

void CESP_NOW::Send_Callback(const uint8_t *mac_addr, esp_now_send_status_t status) {
    //
}

void CESP_NOW::Receive_Callback(const esp_now_recv_info_t *recv_info, const uint8_t *data, int len) {

    CRaw_Data* raw = (CRaw_Data*)malloc(sizeof(CRaw_Data) + len);

    memcpy(raw->src_mac, recv_info->src_addr, ESP_NOW_ETH_ALEN);
    memcpy(raw->dst_mac, recv_info->des_addr, ESP_NOW_ETH_ALEN);
    raw->rssi = recv_info->rx_ctrl->rssi;
    raw->data_len = (uint16_t)len;
    memcpy(&raw->data[0], data, len);

    CLink_Proto_Mgr::Instance()->Incoming_Data(raw);

    free(raw);
}

bool CESP_NOW::Send(const uint8_t* dst_mac, const uint8_t* data, int len) {
    esp_err_t ret = esp_now_send(dst_mac, data, len);

    ESP_LOGI(TAG, "Send returned %i", (int)ret);
    
    return (ret == ESP_OK);
}
