#include "linkproto.h"

#include "freertos/timers.h"
#include "esp_log.h"
#include "esp_random.h"
#include "espnow_layer.h"
#include "../netproto/netproto.h"
#include "esp_timer.h"

#define TAG "LINKPROTO"

CLink_Proto_Mgr* CLink_Proto_Mgr::gInstance;

static constexpr int EVENT_STATE_TRANSITION = BIT0;
static constexpr int EVENT_DATA_FROM_L1 = BIT1;
static constexpr int EVENT_DATA_FROM_L3 = BIT2;
static constexpr int EVENT_INTERNAL = BIT3;
static constexpr int EVENT_EXTERNAL = BIT4;

static constexpr int EVENT_BITS_ALL = EVENT_STATE_TRANSITION | EVENT_DATA_FROM_L1 | EVENT_DATA_FROM_L3 | EVENT_INTERNAL | EVENT_EXTERNAL;

CLink_Proto_Mgr::CLink_Proto_Mgr() : mState(NLink_State::None, this) {

    gInstance = this;

    mEvent_Handle = xEventGroupCreate();
    mIncoming_Data_Queue = xQueueCreate(10, Max_L2_Frame_Size);
    mOutgoing_Data_Queue = xQueueCreate(10, Max_L2_Frame_Size);

    mState.Add_Transition(NLink_State::None, NLink_State::Discovery, &CLink_Proto_Mgr::Transition);
    mState.Add_Transition(NLink_State::Discovery, NLink_State::Operational, &CLink_Proto_Mgr::Transition);

    if (mESP_NOW.Initialize()) {
        mState.Move(NLink_State::Discovery);
    } else {
        ESP_LOGE(TAG, "Could not initialize ESP-NOW");
    }
}

void CLink_Proto_Mgr::Transition(NLink_State from, NLink_State to) {
    ESP_LOGI(TAG, "Transition from %d to %d", (int)from, (int)to);

    if (from == NLink_State::Discovery && to == NLink_State::Operational) {
        CNet_Proto_Mgr::Instance()->Signal_L2_Synchronized();
    }

    xEventGroupSetBits(mEvent_Handle, EVENT_STATE_TRANSITION);
}

void CLink_Proto_Mgr::Incoming_Data(CRaw_Data* data) {
    xQueueSend(mIncoming_Data_Queue, data, 0);
    xEventGroupSetBits(mEvent_Handle, EVENT_DATA_FROM_L1);
}

uint8_t* CLink_Proto_Mgr::Build_Hello_Frame(uint8_t& size) {
    l2pdu::THello* pkt = (l2pdu::THello*)malloc(sizeof(l2pdu::THello));

    pkt->header.opcode = l2pdu::NOpcode::HELLO;
    pkt->magic_number = (uint8_t)esp_random();

    size = sizeof(l2pdu::THello);

    return (uint8_t*)pkt;
}

uint8_t* CLink_Proto_Mgr::Build_Peer_Handshake_Frame(uint8_t& size) {
    l2pdu::TPeer_Handshake* pkt = (l2pdu::TPeer_Handshake*)malloc(sizeof(l2pdu::TPeer_Handshake));

    pkt->header.opcode = l2pdu::NOpcode::HANDSHAKE;
    pkt->peer_count = peers::Get_Peer_Count();

    size = sizeof(l2pdu::TPeer_Handshake);

    return (uint8_t*)pkt;
}

uint8_t* CLink_Proto_Mgr::Build_Peer_Handshake_Reply_Frame(uint8_t& size) {
    l2pdu::TPeer_Handshake_Reply* pkt = (l2pdu::TPeer_Handshake_Reply*)malloc(sizeof(l2pdu::TPeer_Handshake_Reply));

    pkt->header.opcode = l2pdu::NOpcode::HANDSHAKE_REPLY;
    pkt->peer_count = peers::Get_Peer_Count();

    size = sizeof(l2pdu::TPeer_Handshake_Reply);

    return (uint8_t*)pkt;
}

uint8_t* CLink_Proto_Mgr::Build_L3_Data_Frame(uint8_t& size, uint8_t* l3_data, uint8_t l3_data_len) {
    l2pdu::TL3_Data* pkt = (l2pdu::TL3_Data*)malloc(sizeof(l2pdu::TL3_Data) + l3_data_len);

    pkt->header.opcode = l2pdu::NOpcode::L3_PROTO_DATA;
    pkt->data_len = l3_data_len;
    memcpy(&pkt->data[0], l3_data, l3_data_len);

    size = sizeof(l2pdu::TL3_Data) + l3_data_len;

    return (uint8_t*)pkt;
}

void CLink_Proto_Mgr::Send_L3_Data(const uint8_t* dst_mac, uint8_t* data, uint8_t data_len) {

    uint8_t size;
    uint8_t* frame = Build_L3_Data_Frame(size, data, data_len);

    CRaw_Data* raw = (CRaw_Data*)malloc(Max_L2_Frame_Size);
    memcpy(raw->data, frame, size);
    raw->data_len = size;
    memcpy(raw->dst_mac, dst_mac, ESP_NOW_ETH_ALEN);

    xQueueSend(mOutgoing_Data_Queue, raw, 0);
    xEventGroupSetBits(mEvent_Handle, EVENT_DATA_FROM_L3);

    free(frame);
    free(raw);
}

bool CLink_Proto_Mgr::Process_L2_Frame(CRaw_Data* raw) {

    l2pdu::TFrame_Header* hdr = (l2pdu::TFrame_Header*)raw->data;

    uint8_t* reply;
    uint8_t reply_size;

    vTaskDelay((100 + esp_random() % 300) / portTICK_PERIOD_MS);

    switch (hdr->opcode) {
        case l2pdu::NOpcode::NONE:
        default:
            return false;
        case l2pdu::NOpcode::HELLO:
            if (peers::Find_Peer_By_Link_Address(raw->src_mac) == nullptr || mState.Get_Current_State() == NLink_State::Operational) {
                ESP_LOGI(TAG, "Received hello frame, establishing link");
                mESP_NOW.Add_Peer(raw->src_mac);

                reply = Build_Peer_Handshake_Frame(reply_size);
                mESP_NOW.Send(raw->src_mac, reply, reply_size);
            }
            else {
                ESP_LOGI(TAG, "Received hello frame from node that is already a peer");
            }
            break;
        case l2pdu::NOpcode::HANDSHAKE:
            if (peers::Find_Peer_By_Link_Address(raw->src_mac) == nullptr) {
                ESP_LOGI(TAG, "Received handshake frame, establishing link");
                peers::Add_Peer(raw->src_mac);
                mESP_NOW.Add_Peer(raw->src_mac);

                reply = Build_Peer_Handshake_Reply_Frame(reply_size);
                mESP_NOW.Send(raw->src_mac, reply, reply_size);

                // at least one node = operational state
                if (mState.Get_Current_State() == NLink_State::Discovery) {
                    mState.Move(NLink_State::Operational);
                }
            }
            else {
                ESP_LOGI(TAG, "Received handshake frame from node that is already a peer");
            }
            break;
        case l2pdu::NOpcode::HANDSHAKE_REPLY:
            ESP_LOGI(TAG, "Received handshake reply frame, link established!");
            peers::Add_Peer(raw->src_mac);

            // at least one node = operational state
            if (mState.Get_Current_State() == NLink_State::Discovery) {
                mState.Move(NLink_State::Operational);
            }
            break;
        case l2pdu::NOpcode::L3_PROTO_DATA:
            if (mState.Get_Current_State() != NLink_State::Operational) {
                ESP_LOGI(TAG, "Received L3 data during non-operational state, discarding");
                return false;
            }

            l2pdu::TL3_Data* dataptr = (l2pdu::TL3_Data*)hdr;

            CData_Container* cont = (CData_Container*)&dataptr->data_len;
            CNet_Proto_Mgr::Instance()->Incoming_Data(cont);
            break;
    }

    return true;
}

void CLink_Proto_Mgr::Task_Run() {

    ESP_LOGI(TAG, "Task running");

    uint32_t lastUpdate = esp_timer_get_time() / 1000ULL;

    while (true) {

        EventBits_t bits = xEventGroupWaitBits(mEvent_Handle, EVENT_BITS_ALL, pdFALSE, pdFALSE, (3000 + (esp_random() % 500)) / portTICK_PERIOD_MS);

        uint32_t curTime = (esp_timer_get_time() / 1000ULL);

        if (bits == 0 || curTime - lastUpdate > 3000) {

            lastUpdate = curTime;

            if (mState.Get_Current_State() == NLink_State::Discovery) {
                uint8_t size;
                uint8_t* hello = Build_Hello_Frame(size);

                bool succ = mESP_NOW.Send(Broadcast_L2_Address, hello, size);
                ESP_LOGI(TAG, "Broadcasting... %d", (int)succ);
                continue;
            }
        }

        ESP_LOGI(TAG, "Triggered event bitmask: %d", (int)bits);

        if (bits & EVENT_STATE_TRANSITION) {
            //
        }

        if (bits & EVENT_DATA_FROM_L1) {
            CRaw_Data* data = (CRaw_Data*)malloc(Max_L2_Frame_Size);

            while (xQueueReceive(mIncoming_Data_Queue, data, 0)) {
                ESP_LOGI(TAG, "Received a frame from " MACSTR " for " MACSTR " with RSSI %i", MAC2STR(data->src_mac), MAC2STR(data->dst_mac), data->rssi);

                Process_L2_Frame(data);
            }

            free(data);
        }

        if (bits & EVENT_DATA_FROM_L3) {
            CRaw_Data* data = (CRaw_Data*)malloc(Max_L2_Frame_Size);

            while (xQueueReceive(mOutgoing_Data_Queue, data, 0)) {
                ESP_LOGI(TAG, "Send frame to " MACSTR, MAC2STR(data->dst_mac));

                mESP_NOW.Send(data->dst_mac, data->data, data->data_len);
            }

            free(data);
        }

        if (bits & EVENT_INTERNAL) {
            //
        }

        if (bits & EVENT_EXTERNAL) {
            //
        }

        xEventGroupClearBits(mEvent_Handle, 0xFF);
    }
}
