#ifndef BSS_NETPROTO_H
#define BSS_NETPROTO_H

#include "freertos/FreeRTOS.h"

#include "../common/statemachine.h"
#include "../common/taskrunnable.h"
#include "freertos/event_groups.h"
#include "freertos/queue.h"

#include "../linkproto/peertable.h"

enum class NNet_State {
    None,
    Master_Discovery,
    Address_Assignment,
    Operational,

    count
};

#pragma pack(push, 1)

struct CData_Container {
    uint16_t data_len;
    uint8_t data[0];    // just a data break
};

#pragma pack(pop)

static constexpr uint8_t Invalid_L3_Address = 0x00;
static constexpr uint8_t Master_L3_Address = 0x01;
static constexpr uint8_t Broadcast_L3_Address = 0xFF;

static constexpr int L3_Master_Discovery_Attempts = 10;
static constexpr int L3_Address_Assignment_Ticks_Skip = 5;
static constexpr uint8_t L3_Default_TTL = 5;

namespace l3pdu {

    enum class NOpcode : uint8_t {
        NONE,
        NETADDR_MASTER_DISCOVERY,
        NETADDR_MASTER_DISCOVERY_REPLY,
        NETADDR_ADDR_REQUEST,
        NETADDR_ADDR_OFFER,
        NETADDR_ADDR_ACK,
    };

#pragma pack(push, 1)

    struct THeader {
        NOpcode opcode;
        uint8_t src_addr;
        uint8_t dst_addr;
        uint8_t ttl;
    };

    struct TNetaddr_Master_Discovery {
        THeader header;
        uint8_t my_mac[ESP_NOW_ETH_ALEN];
    };

    struct TNetaddr_Master_Discovery_Reply {
        THeader header;
        uint8_t current_master_mac[ESP_NOW_ETH_ALEN];
    };

    struct TNetaddr_Addr_Request {
        THeader header;
        uint8_t my_mac[ESP_NOW_ETH_ALEN];
    };

    struct TNetaddr_Addr_Offer {
        THeader header;
        uint8_t addr;
    };

    struct TNetaddr_Addr_Ack {
        THeader header;
        uint8_t addr;
    };

#pragma pack(pop)
}

class CNet_Proto_Mgr : public CTask_Runnable {
    private:
        CState_Machine<NNet_State, CNet_Proto_Mgr> mState;
        EventGroupHandle_t mEvent_Handle;
        QueueHandle_t mIncoming_Data_Queue;
        QueueHandle_t mOutgoing_Data_Queue;

        static CNet_Proto_Mgr* gInstance;

        int mRemaining_Discovery_Ticks = 0;
        int mRemaining_Address_Assignment_Ticks = 0;

        uint8_t mLast_Assigned_Address = 0;

        uint8_t mMy_Address = Invalid_L3_Address;

        uint8_t mCurrent_Master_L2_Addr[ESP_NOW_ETH_ALEN];

    protected:
        void Transition(NNet_State from, NNet_State to);

        void Send_Master_Discovery();
        void Send_Master_Discovery_Reply(uint8_t* master_mac);
        void Send_Netaddr_Assignment_Request();
        void Send_Netaddr_Assignment_Offer(uint8_t addr);
        void Send_Netaddr_Assignment_Ack();

        void Process_L3_Packet(CData_Container* raw);

        const uint8_t* Select_L2_Address(uint8_t l3_addr);

    public:
        CNet_Proto_Mgr();

        void Task_Run() override;
        void Incoming_Data(CData_Container* data);

        void Signal_L2_Synchronized();

        static CNet_Proto_Mgr* Instance() {
            return gInstance;
        }
};

#endif
