#include "netproto.h"

#include "freertos/timers.h"
#include "esp_log.h"

#include "../linkproto/linkproto.h"

#define TAG "NETPROTO"

CNet_Proto_Mgr* CNet_Proto_Mgr::gInstance;

static constexpr int EVENT_STATE_TRANSITION = BIT0;
static constexpr int EVENT_DATA_FROM_L2 = BIT1;
static constexpr int EVENT_DATA_FROM_L4 = BIT2;
static constexpr int EVENT_INTERNAL = BIT3;
static constexpr int EVENT_EXTERNAL = BIT4;

static constexpr int EVENT_BITS_ALL = EVENT_STATE_TRANSITION | EVENT_DATA_FROM_L2 | EVENT_DATA_FROM_L4 | EVENT_INTERNAL | EVENT_EXTERNAL;

CNet_Proto_Mgr::CNet_Proto_Mgr() : mState(NNet_State::None, this) {

    gInstance = this;
    mMy_Address = Invalid_L3_Address;

    mEvent_Handle = xEventGroupCreate();
    mIncoming_Data_Queue = xQueueCreate(10, Max_L2_Frame_Size);
    mOutgoing_Data_Queue = xQueueCreate(10, Max_L2_Frame_Size);

    mState.Add_Transition(NNet_State::None, NNet_State::Master_Discovery, &CNet_Proto_Mgr::Transition);
    mState.Add_Transition(NNet_State::Master_Discovery, NNet_State::Address_Assignment, &CNet_Proto_Mgr::Transition);
    mState.Add_Transition(NNet_State::Address_Assignment, NNet_State::Operational, &CNet_Proto_Mgr::Transition);
}

void CNet_Proto_Mgr::Transition(NNet_State from, NNet_State to) {
    ESP_LOGI(TAG, "Transition from %d to %d", (int)from, (int)to);

    if (to == NNet_State::Master_Discovery) {
        mRemaining_Discovery_Ticks = L3_Master_Discovery_Attempts;
        mMy_Address = Invalid_L3_Address;
    }
    else if (to == NNet_State::Address_Assignment) {
        mMy_Address = Invalid_L3_Address;
    }

    xEventGroupSetBits(mEvent_Handle, EVENT_STATE_TRANSITION);
}

void CNet_Proto_Mgr::Incoming_Data(CData_Container* data) {
    xQueueSend(mIncoming_Data_Queue, data, 0);
    xEventGroupSetBits(mEvent_Handle, EVENT_DATA_FROM_L2);
}

void CNet_Proto_Mgr::Signal_L2_Synchronized() {
    if (mState.Get_Current_State() == NNet_State::None) {
        esp_base_mac_addr_get(mCurrent_Master_L2_Addr);
        mState.Move(NNet_State::Master_Discovery);
    }
}

void CNet_Proto_Mgr::Send_Master_Discovery() {

    l3pdu::TNetaddr_Master_Discovery* packet = new l3pdu::TNetaddr_Master_Discovery;

    packet->header.opcode = l3pdu::NOpcode::NETADDR_MASTER_DISCOVERY;
    packet->header.src_addr = Invalid_L3_Address;
    packet->header.dst_addr = Broadcast_L3_Address;
    packet->header.ttl = L3_Default_TTL;
    
    esp_base_mac_addr_get(packet->my_mac);

    CLink_Proto_Mgr::Instance()->Send_L3_Data(Broadcast_L2_Address, (uint8_t*)packet, sizeof(l3pdu::TNetaddr_Master_Discovery));
}

void CNet_Proto_Mgr::Send_Master_Discovery_Reply(uint8_t* master_mac) {

    l3pdu::TNetaddr_Master_Discovery_Reply* packet = new l3pdu::TNetaddr_Master_Discovery_Reply;

    packet->header.opcode = l3pdu::NOpcode::NETADDR_MASTER_DISCOVERY_REPLY;
    packet->header.src_addr = Invalid_L3_Address;
    packet->header.dst_addr = Broadcast_L3_Address;
    packet->header.ttl = L3_Default_TTL;
    
    memcpy(packet->current_master_mac, master_mac, ESP_NOW_ETH_ALEN);

    CLink_Proto_Mgr::Instance()->Send_L3_Data(Broadcast_L2_Address, (uint8_t*)packet, sizeof(l3pdu::TNetaddr_Master_Discovery_Reply));
}

void CNet_Proto_Mgr::Send_Netaddr_Assignment_Request() {

    l3pdu::TNetaddr_Addr_Request* packet = new l3pdu::TNetaddr_Addr_Request;

    packet->header.opcode = l3pdu::NOpcode::NETADDR_ADDR_REQUEST;
    packet->header.src_addr = Invalid_L3_Address;
    packet->header.dst_addr = Master_L3_Address;
    packet->header.ttl = L3_Default_TTL;

    const uint8_t* target_addr = Select_L2_Address(packet->header.dst_addr);
    if (!target_addr) {
        return;
    }

    esp_base_mac_addr_get(packet->my_mac);

    CLink_Proto_Mgr::Instance()->Send_L3_Data(target_addr, (uint8_t*)packet, sizeof(l3pdu::TNetaddr_Addr_Request));
}

void CNet_Proto_Mgr::Send_Netaddr_Assignment_Offer(uint8_t addr) {
    l3pdu::TNetaddr_Addr_Offer* packet = new l3pdu::TNetaddr_Addr_Offer;

    packet->header.opcode = l3pdu::NOpcode::NETADDR_ADDR_OFFER;
    packet->header.src_addr = mMy_Address;
    packet->header.dst_addr = Broadcast_L3_Address;
    packet->header.ttl = L3_Default_TTL;

    packet->addr = addr;

    const uint8_t* target_addr = Select_L2_Address(packet->header.dst_addr);
    if (!target_addr) {
        return;
    }

    CLink_Proto_Mgr::Instance()->Send_L3_Data(target_addr, (uint8_t*)packet, sizeof(l3pdu::TNetaddr_Addr_Offer));
}

void CNet_Proto_Mgr::Send_Netaddr_Assignment_Ack() {
    l3pdu::TNetaddr_Addr_Ack* packet = new l3pdu::TNetaddr_Addr_Ack;

    packet->header.opcode = l3pdu::NOpcode::NETADDR_ADDR_ACK;
    packet->header.src_addr = mMy_Address;
    packet->header.dst_addr = Master_L3_Address;
    packet->header.ttl = L3_Default_TTL;

    packet->addr = mMy_Address;

    const uint8_t* target_addr = Select_L2_Address(packet->header.dst_addr);
    if (!target_addr) {
        return;
    }

    CLink_Proto_Mgr::Instance()->Send_L3_Data(target_addr, (uint8_t*)packet, sizeof(l3pdu::TNetaddr_Addr_Ack));
}

void CNet_Proto_Mgr::Send_L4_Data(uint8_t dst_addr, COut_Data_Container* raw) {
    l3pdu::TL4_Data* packet = (l3pdu::TL4_Data*)malloc(sizeof(l3pdu::TL4_Data) + raw->data_len);

    packet->header.opcode = l3pdu::NOpcode::L4_DATA;
    packet->header.src_addr = mMy_Address;
    packet->header.dst_addr = dst_addr;
    packet->header.ttl = L3_Default_TTL;

    packet->data_len = raw->data_len;

    memcpy(&packet->data[0], raw->data, raw->data_len);

    const uint8_t* target_addr = Select_L2_Address(packet->header.dst_addr);
    if (!target_addr) {
        return;
    }

    CLink_Proto_Mgr::Instance()->Send_L3_Data(target_addr, (uint8_t*)packet, sizeof(l3pdu::TL4_Data) + raw->data_len);
}

void CNet_Proto_Mgr::Process_L3_Packet(CData_Container* raw) {

    l3pdu::THeader* header = (l3pdu::THeader*)raw->data;

    ESP_LOGI(TAG, "Opcode = %i, dest = %i", (int)header->opcode, (int)header->dst_addr);

    // routing! (only if in operational state)
    if (header->dst_addr != mMy_Address && header->dst_addr != Broadcast_L3_Address) {

        if (mState.Get_Current_State() == NNet_State::Operational) {
            ESP_LOGW(TAG, "Routing data (my addr = %i) for address %i", (int)mMy_Address, (int)header->dst_addr);

            const uint8_t* target_addr = Select_L2_Address(header->dst_addr);
            if (!target_addr) {
                return;
            }

            if (header->ttl > 0) {
                header->ttl--;
                CLink_Proto_Mgr::Instance()->Send_L3_Data(target_addr, (uint8_t*)raw->data, raw->data_len);
            }
            return;
        }
        else {
            return;
        }
    }

    switch (header->opcode) {
        case l3pdu::NOpcode::NONE:
        default:
            break;

        case l3pdu::NOpcode::NETADDR_MASTER_DISCOVERY:
        {
            l3pdu::TNetaddr_Master_Discovery* disc = (l3pdu::TNetaddr_Master_Discovery*)header;

            // elect master only if we are in discovery state
            if (mState.Get_Current_State() == NNet_State::Master_Discovery) {

                bool incoming_greater = false;
                for (int i = 0; i < ESP_NOW_ETH_ALEN; i++) {
                    if (disc->my_mac[i] > mCurrent_Master_L2_Addr[i]) {
                        ESP_LOGI(TAG, "%02x > %02x", (int)disc->my_mac[i], (int)mCurrent_Master_L2_Addr[i]);
                        incoming_greater = true;
                        break;
                    }
                    else if (disc->my_mac[i] < mCurrent_Master_L2_Addr[i]) {
                        break;
                    }
                }

                if (incoming_greater) {
                    ESP_LOGI(TAG, "Old master: " MACSTR ", new master " MACSTR, MAC2STR(mCurrent_Master_L2_Addr), MAC2STR(disc->my_mac));
                    memcpy(mCurrent_Master_L2_Addr, disc->my_mac, ESP_NOW_ETH_ALEN);
                }

            }

            ESP_LOGI(TAG, "Send master discovery reply");
            Send_Master_Discovery_Reply(mCurrent_Master_L2_Addr);
            break;
        }

        case l3pdu::NOpcode::NETADDR_MASTER_DISCOVERY_REPLY:
        {
            l3pdu::TNetaddr_Master_Discovery_Reply* disc = (l3pdu::TNetaddr_Master_Discovery_Reply*)header;

            ESP_LOGI(TAG, "Received master discovery reply, current master is " MACSTR, MAC2STR(disc->current_master_mac));

            memcpy(mCurrent_Master_L2_Addr, disc->current_master_mac, ESP_NOW_ETH_ALEN);
            break;
        }
        case l3pdu::NOpcode::NETADDR_ADDR_REQUEST:
        {
            l3pdu::TNetaddr_Addr_Request* disc = (l3pdu::TNetaddr_Addr_Request*)header;

            ESP_LOGI(TAG, "Received address request");

            // TODO: this is not ideal, as it does not account for the ACKs
            mLast_Assigned_Address++;
            uint8_t addr = mLast_Assigned_Address;

            const TPeer_Table_Entry* peer = peers::Find_Peer_By_Link_Address(disc->my_mac);
            if (peer) {
                peers::Set_Peer_Net_Address(peer->link_addr, addr);
            }

            Send_Netaddr_Assignment_Offer(addr);
            break;
        }
        case l3pdu::NOpcode::NETADDR_ADDR_OFFER:
        {
            l3pdu::TNetaddr_Addr_Offer* disc = (l3pdu::TNetaddr_Addr_Offer*)header;
            ESP_LOGI(TAG, "Master assigned address = %i", (int)disc->addr);
            mMy_Address = disc->addr;
            Send_Netaddr_Assignment_Ack();
            mState.Move(NNet_State::Operational);
            break;
        }
        case l3pdu::NOpcode::NETADDR_ADDR_ACK:
        {
            ESP_LOGI(TAG, "Received address ack");
            break;
        }
        case l3pdu::NOpcode::L4_DATA:
        {
            l3pdu::TL4_Data* d = (l3pdu::TL4_Data*)header;

            ESP_LOGI(TAG, "Incoming L4 data len = %i", (int)d->data_len);

            CSensor_Mgr::Instance()->Incoming_Data(d->data, d->data_len);
            break;
        }
    }
}

const uint8_t* CNet_Proto_Mgr::Select_L2_Address(uint8_t l3_addr) {

    const TPeer_Table_Entry* target_peer = peers::Find_Peer_By_Net_Address(l3_addr);
    if (target_peer) {
        return target_peer->link_addr;
    }

    // NOTE: this is the point, where you implement the routing mechanism
    // for now, we will implement rumor routing - select random peer

    target_peer = peers::Select_Random_Peer();
    if (!target_peer) {
        return nullptr;
    }

    return target_peer->link_addr;
}

void CNet_Proto_Mgr::Send_Data(uint8_t dest_addr, uint8_t* data, uint16_t data_len) {

    COut_Data_Container* odc =  (COut_Data_Container*)malloc(sizeof(COut_Data_Container) + data_len);

    if (!odc) {
        return;
    }

    odc->dest_addr = dest_addr;
    odc->data_len = data_len;
    memcpy(&odc->data[0], data, data_len);

    xQueueSend(mOutgoing_Data_Queue, odc, 0);
    xEventGroupSetBits(mEvent_Handle, EVENT_DATA_FROM_L4);
}

void CNet_Proto_Mgr::Task_Run() {

    ESP_LOGI(TAG, "Task running");

    while (true) {

        EventBits_t bits = xEventGroupWaitBits(mEvent_Handle, EVENT_BITS_ALL, pdFALSE, pdFALSE, 2000 / portTICK_PERIOD_MS);

        if (bits == 0) {
            if (mState.Get_Current_State() == NNet_State::Master_Discovery) {
                if (mRemaining_Discovery_Ticks == 0) {
                    mState.Move(NNet_State::Address_Assignment);
                    continue;
                }
                else {
                    ESP_LOGI(TAG, "Send master discovery");
                    mRemaining_Discovery_Ticks--;
                    Send_Master_Discovery();
                }
            }
            else if (mState.Get_Current_State() == NNet_State::Address_Assignment) {
                ESP_LOGI(TAG, "Ticks to address request: %i", (int)mRemaining_Address_Assignment_Ticks);
                if (mRemaining_Address_Assignment_Ticks == 0) {
                    mRemaining_Address_Assignment_Ticks = L3_Address_Assignment_Ticks_Skip;

                    Send_Netaddr_Assignment_Request();
                }
                else {
                    mRemaining_Address_Assignment_Ticks--;
                }
            }
            else {
                continue;
            }
        }

        ESP_LOGI(TAG, "Triggered event bitmask: %d", (int)bits);

        if (bits & EVENT_STATE_TRANSITION) {
            if (mState.Get_Current_State() == NNet_State::Address_Assignment) {

                uint8_t my_mac[ESP_NOW_ETH_ALEN];
                esp_base_mac_addr_get(my_mac);

                if (memcmp(mCurrent_Master_L2_Addr, my_mac, ESP_NOW_ETH_ALEN) == 0) {
                    mMy_Address = Master_L3_Address;
                    mLast_Assigned_Address = Master_L3_Address;
                    mState.Move(NNet_State::Operational);
                }
                else {
                    mRemaining_Address_Assignment_Ticks = L3_Address_Assignment_Ticks_Skip;

                    const TPeer_Table_Entry* master_peer = peers::Find_Peer_By_Link_Address(mCurrent_Master_L2_Addr);
                    if (master_peer) {
                        peers::Set_Peer_Net_Address(mCurrent_Master_L2_Addr, Master_L3_Address);
                    }
                }
            }
        }

        if (bits & EVENT_DATA_FROM_L2) {
            CData_Container* data = (CData_Container*)malloc(Max_L2_Frame_Size);

            while (xQueueReceive(mIncoming_Data_Queue, data, 0)) {
                ESP_LOGI(TAG, "Received a L3 packet");

                Process_L3_Packet(data);
            }

            free(data);
        }

        if (bits & EVENT_DATA_FROM_L4) {
            COut_Data_Container* data = (COut_Data_Container*)malloc(Max_L2_Frame_Size);

            while (xQueueReceive(mOutgoing_Data_Queue, data, 0)) {
                ESP_LOGI(TAG, "Sending a higher layer packet");

                Send_L4_Data(data->dest_addr, data);
            }

            free(data);
        }

        if (bits & EVENT_INTERNAL) {
            //
        }

        if (bits & EVENT_EXTERNAL) {
            //
        }

        xEventGroupClearBits(mEvent_Handle, 0xFF);
    }
}
